unit UnitMD5;
{
Unit: UnitMD5
By: Stucuk
Description: This Unit Loads MD5 Files (Doom 3 Model Format)

Spcial Thanks to Vipr for all his help in making me understand the md5 format.

}


interface

//mesh
{
	// meshes: RLFX1
	shader "models/weapons/rocketlauncher/rocketlauncher_fx"

	numverts 4
           id  u v   wid  num weight
	vert 0 ( 0 0 ) 2 1
	vert 1 ( 0.9999998808 0 ) 1 1
	vert 2 ( 1 1 ) 0 1
	vert 3 ( 0 1 ) 3 1

	numtris 2
         id  vert id's
	tri 0 2 1 0
	tri 1 3 2 0

	numweights 4
             id bID  ?                  POSITION
	weight 0 103 1 ( -7.7185444832 -5.7600746155 0.069297567 )
	weight 1 103 1 ( -5.494342804 7.6056361198 0.6600854993 )
	weight 2 103 1 ( 7.7353591919 5.4382991791 -0.1142965928 )
	weight 3 103 1 ( 5.5111575127 -7.9274125099 -0.7050862312 )
}

// to get 4th quaternion
//a^2+b^2+c^2+d^2 = 1
//a^2 = 1-b^2-c^2-d^2
//a = Sqr(1-b^2-c^2-d^2)

{

QuatPtr=^Quat;
 Quat=RECORD
  w,x,y,z:single;
 end;


function Q2M4(a:QuatPtr):PGLMatrixf;   // quaternion to matrix 4x4
var x2,y2,z2,xy,wx,wy,wz,xz,yz:single;
begin
 result:=new(PGLMatrixf);
 x2:=2*sqr(a.x); y2:=2*sqr(a.y); z2:=2*sqr(a.z);
 xy:=2*a.x*a.y;  wz:=2*a.w*a.z;  yz:=2*a.y*a.z;
 xz:=2*a.x*a.z;  wy:=2*a.w*a.y;  wx:=2*a.w*a.x;
 result[1,1]:=1-y2-z2; result[2,1]:=xy+wz;   result[3,1]:=xz-wz;   result[4,1]:=0;
 result[1,2]:=xy-wz;   result[2,2]:=1-x2-z2; result[3,2]:=yz-wx;   result[4,2]:=0;
 result[1,3]:=xz-wy;   result[2,3]:=yz-wx;   result[3,3]:=1-x2-y2; result[4,3]:=0;
 result[1,4]:=0;       result[2,4]:=0;       result[3,4]:=0;       result[4,4]:=1;
end;



}

Uses Windows,SysUtils,Geometry,Math3d,Classes, Helios_Global,OpenGl15,Helios_Types;

Type
TQ = record
X,Y,Z,W : Single;
end;

TMD5Joint = record
Name : String;
Parent : integer;
Position,Quaternion : TVector3f;
Matrix : TMatrix;
end;

TMD5Vert = record
s,t : Single;
WeightId,WeightCount : Integer;
Position : TVector3f; // Generated At Runtime
Colour : TVector3f;
end;

TMD5Tri = record
V1,V2,V3 : Integer;
end;

TMD5Weight = record
BoneId : Integer;
Factor : Single;
Position : TVector3f;
end;

TMD5MeshHeader = record
NumVerts,NumTris,NumWeights : integer;
end;

TMD5Mesh = record
Shader : String;
ShaderID : Integer; // Generated At Runtime
Header : TMD5MeshHeader;
Verts : Array Of TMD5Vert;
Tris : Array Of TMD5Tri;
Weights : Array Of TMD5Weight;
end;

TMD5Header = record
MD5Version,NumJoints,NumMeshes : integer;
end;

TMD5File = record
Header : TMD5Header;
Joints : Array Of TMD5Joint;
Meshs : Array Of TMD5Mesh;
end;

Function LoadMD5(Var MD5 : TMD5File; Const Filename : String) : Boolean;
Procedure BuildMeshs(Var MD5 : TMD5File);
Procedure RenderMD5(Var MD5 : TMD5File);
Procedure RenderMD5Mesh(Var Mesh : TMD5Mesh);

implementation

function Q2M4(a:Tq):TMatrix;   // quaternion to matrix 4x4
var x2,y2,z2,xy,wx,wy,wz,xz,yz:single;
begin
 x2:=2*sqr(a.x); y2:=2*sqr(a.y); z2:=2*sqr(a.z);
 xy:=2*a.x*a.y;  wz:=2*a.w*a.z;  yz:=2*a.y*a.z;
 xz:=2*a.x*a.z;  wy:=2*a.w*a.y;  wx:=2*a.w*a.x;
 result[0,0]:=1-y2-z2; result[1,0]:=xy+wz;   result[2,0]:=xz-wz;   result[3,0]:=0;
 result[0,1]:=xy-wz;   result[1,1]:=1-x2-z2; result[2,1]:=yz-wx;   result[3,1]:=0;
 result[0,2]:=xz-wy;   result[1,2]:=yz-wx;   result[2,2]:=1-x2-y2; result[3,2]:=0;
 result[0,3]:=0;       result[1,3]:=0;       result[2,3]:=0;       result[3,3]:=1;
end;

Function GetValue(Line : Integer; StringList : TStringList) : String;
Var
First,Rest : String;
begin

GetFirstBit(FindStart(StringList.Strings[Line]),First,Rest);

GetFirstBit(FindStart(Rest),First,Rest);

Result := First;
end;

Function GetFirstValue(Line : Integer; StringList : TStringList) : String;
Var
First,Rest : String;
begin

GetFirstBit(FindStart(StringList.Strings[Line]),First,Rest);

Result := First;
end;

Function ReturnValue(Find : String; LineStart : Integer; StringList : TStringList; Var Line : Integer) : Boolean;
var
S : String;
Begin

Line := LineStart-1;
Result := False;

repeat
Inc(Line);

If Line >= StringList.Count then
exit;

S := GetFirstValue(Line,StringList);

until Ansilowercase(S) = ansilowercase(Find);

Result := True;

end;

Function ReturnValue2(Find : String; LineStart : Integer; StringList : TStringList; Var Line : Integer) : Boolean;
var
S : String;
FF,FB : Boolean;
Begin

Line := LineStart-1;
Result := False;
FF := False;
FB := False;

repeat
Inc(Line);

If Line >= StringList.Count then
exit;

S := GetFirstValue(Line,StringList);

If Ansilowercase(s) = Ansilowercase(Find) then
FF := True;

If FF then
If Pos('{',StringList.Strings[Line]) > 0 then
FB := True;

until (FF and FB);

Result := True;

end;

Function LoadJoints(Var Line : Integer; Var MD5File : TMD5File; Const StringList : TStringList) : Boolean;
var
S,First,Rest : String;
JId : Integer;
Q : TQ;
Term : Single;
begin

Result := False;

JID := 0;

Repeat
Inc(Line);

If Line >= StringList.Count then
exit;

S := GetFirstValue(Line,StringList);

If Copy(S,1,1) = '"' then
begin
GetFirstBit(FindStart(StringReplace(StringList.Strings[Line], #9,  ' ', [rfReplaceAll])),First,Rest);
MD5File.Joints[JID].Name := StringReplace(First, '"',  '', [rfReplaceAll]);
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Joints[JID].Parent := StrToInt(First);
GetFirstBit(FindStart(Rest),First,Rest);// (
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Joints[JID].Position.X := StrToFloat(First);
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Joints[JID].Position.Y := StrToFloat(First);
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Joints[JID].Position.Z := StrToFloat(First);

GetFirstBit(FindStart(Rest),First,Rest); // )
GetFirstBit(FindStart(Rest),First,Rest); // (

GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Joints[JID].Quaternion.X := StrToFloat(First);
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Joints[JID].Quaternion.Y := StrToFloat(First);
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Joints[JID].Quaternion.Z := StrToFloat(First);
Q.X := MD5File.Joints[JID].Quaternion.X;
Q.Y := MD5File.Joints[JID].Quaternion.Y;
Q.Z := MD5File.Joints[JID].Quaternion.Z;

Term := 1.0 - (Q.X*Q.X) - (Q.Y*Q.Y) - (Q.Z*Q.Z);
if (term < 0) then
   Q.W := 0
else
   Q.W := - sqrt(term);

MD5File.Joints[JID].Matrix := Q2M4(Q);
Inc(JID);
end;

Until  S = '}';

Result := True;

end;

Function LoadMesh(Var Line,MeshNo : Integer; Var MD5File : TMD5File; Const StringList : TStringList) : Boolean;
var
S,First,Rest : String;
VID,TID,WID : Integer;
begin

Result := False;

VID := 0;
TID := 0;
WID := 0;

Repeat
Inc(Line);

If Line >= StringList.Count then
exit;

S := ansilowercase(GetFirstValue(Line,StringList));

If S = 'shader' then
begin
MD5File.Meshs[MeshNo].Shader := StringReplace(GetValue(Line,StringList), '"',  '', [rfReplaceAll]);
MD5File.Meshs[MeshNo].ShaderID := -2;
ShaderManager.AddToShaderManager(MD5File.Meshs[MeshNo].Shader);
end
else
If S = 'numverts' then
begin
MD5File.Meshs[MeshNo].Header.NumVerts := StrtoInt(GetValue(Line,StringList));
SetLength(MD5File.Meshs[MeshNo].Verts,MD5File.Meshs[MeshNo].Header.NumVerts);
VID := 0;
end
else
If S = 'vert' then
begin
GetFirstBit(FindStart(StringReplace(StringList.Strings[Line], #9,  ' ', [rfReplaceAll])),First,Rest);
GetFirstBit(FindStart(Rest),First,Rest);
VID := StrToInt(First);
GetFirstBit(FindStart(Rest),First,Rest); // (
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Meshs[MeshNo].Verts[VID].s := StrToFloat(First);
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Meshs[MeshNo].Verts[VID].t := StrToFloat(First);

GetFirstBit(FindStart(Rest),First,Rest); // )
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Meshs[MeshNo].Verts[VID].WeightId := StrToInt(First);
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Meshs[MeshNo].Verts[VID].WeightCount := StrToInt(First);
Inc(VID);
end
else
If S = 'numtris' then
begin
MD5File.Meshs[MeshNo].Header.NumTris := StrtoInt(GetValue(Line,StringList));
SetLength(MD5File.Meshs[MeshNo].Tris,MD5File.Meshs[MeshNo].Header.NumTris);
TID := 0;
end
else
If S = 'tri' then
begin
GetFirstBit(FindStart(StringReplace(StringList.Strings[Line], #9,  ' ', [rfReplaceAll])),First,Rest);
GetFirstBit(FindStart(Rest),First,Rest);
TID := StrToInt(First);
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Meshs[MeshNo].Tris[TID].V1 := StrToInt(First);
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Meshs[MeshNo].Tris[TID].V2 := StrToInt(First);
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Meshs[MeshNo].Tris[TID].V3 := StrToInt(First);
Inc(TID);
end
else
If S = 'numweights' then
begin
MD5File.Meshs[MeshNo].Header.NumWeights := StrtoInt(GetValue(Line,StringList));
SetLength(MD5File.Meshs[MeshNo].Weights,MD5File.Meshs[MeshNo].Header.NumWeights);
WID := 0;
end
else
If S = 'weight' then
begin
GetFirstBit(FindStart(StringReplace(StringList.Strings[Line], #9,  ' ', [rfReplaceAll])),First,Rest);
GetFirstBit(FindStart(Rest),First,Rest);
WID := StrToInt(First);
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Meshs[MeshNo].Weights[WID].BoneId := StrToInt(First);
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Meshs[MeshNo].Weights[WID].Factor := StrToFloat(First);
GetFirstBit(FindStart(Rest),First,Rest); // (
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Meshs[MeshNo].Weights[WID].Position.X := StrToFloat(First);
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Meshs[MeshNo].Weights[WID].Position.Y := StrToFloat(First);
GetFirstBit(FindStart(Rest),First,Rest);
MD5File.Meshs[MeshNo].Weights[WID].Position.Z := StrToFloat(First);
Inc(WID);
end;

Until  S = '}';
Inc(MeshNo);

Result := True;

end;

Function LoadMD5(Var MD5 : TMD5File; Const Filename : String) : Boolean;
var
S : TStringList;
Line : Integer;
MeshNo,x : Integer;
begin
Result := False;
If Not Fileexists(Filename) then Exit;

S := TStringList.Create;
S.LoadFromFile(Filename);

Line := 0;

If not ReturnValue('md5version',Line,S,Line) then
begin
Messagebox(0,'Error: MD5Version Not Found','MD5 Load Error',0);
Exit;
end;

MD5.Header.MD5Version := StrToInt(GetValue(Line,S));

If not ReturnValue('numjoints',Line,S,Line) then
begin
Messagebox(0,'Error: NumJoints Not Found','MD5 Load Error',0);
Exit;
end;

MD5.Header.NumJoints := StrToInt(GetValue(Line,S));
SetLength(MD5.Joints,MD5.Header.NumJoints);

If not ReturnValue('nummeshes',Line,S,Line) then
begin
Messagebox(0,'Error: NumMeshes Not Found','MD5 Load Error',0);
Exit;
end;

MD5.Header.NumMeshes := StrToInt(GetValue(Line,S));
SetLength(MD5.Meshs,MD5.Header.NumMeshes);

If not ReturnValue('joints',Line,S,Line) then
begin
Messagebox(0,'Error: Joints Not Found','MD5 Load Error',0);
Exit;
end;

If not LoadJoints(Line,MD5,S) then
begin
Messagebox(0,'Error: Loading Joints','MD5 Load Error',0);
Exit;
end;

MeshNo := 0;

For x := 0 to MD5.Header.NumMeshes-1 do
If not LoadMesh(Line,MeshNo,MD5,S) then
begin
Messagebox(0,'Error: Loading Joints','MD5 Load Error',0);
Exit;
end;

BuildMeshs(MD5);

Messagebox(0,pchar('Version: ' + inttostr(MD5.Header.MD5Version) + #13 + 'NumJoints: ' + inttostr(MD5.Header.NumJoints) + #13 + 'NumMeshes: ' + inttostr(MD5.Header.NumMeshes) + #13),'MD5Info',0);

S.Free;
Result := True;
end;

Function MakeTransMatrix(Vector : TVector3f) : TMatrix;
begin

Result[0,0] := 1;
Result[0,1] := 0;
Result[0,2] := 0;

Result[1,0] := 0;
Result[1,1] := 1;
Result[1,2] := 0;

Result[2,0] := 0;
Result[2,1] := 0;
Result[2,2] := 1;

Result[3,0] := 0;
Result[3,1] := 0;
Result[3,2] := 0;
Result[3,3] := 1;

Result[0,3] := Vector.X;
Result[1,3] := Vector.Y;
Result[2,3] := Vector.Z;

end;

Function GetPosWithMatrix(var Pos : TVector3f; Const Matrix : TMatrix) : TVector3f;
begin
  Result.X := ( Pos.x * Matrix[0,0] + Pos.y * Matrix[0,1] + Pos.z * Matrix[0,2] + Matrix[0,3]);
  Result.Y := ( Pos.x * Matrix[1,0] + Pos.y * Matrix[1,1] + Pos.z * Matrix[1,2] + Matrix[1,3]);
  Result.Z := ( Pos.x * Matrix[2,0] + Pos.y * Matrix[2,1] + Pos.z * Matrix[2,2] + Matrix[2,3]);
end;

Procedure BuildMD5Mesh(MD5 : TMD5File; Mesh : TMD5Mesh);
var
V,W : Integer;
Pos,Joint : TVector3f;
TM : TMatrix;
Weighting : Single;
begin

// Build The Vertex's
For V := 0 to Mesh.Header.NumVerts-1 do
begin
Mesh.Verts[v].Position := SetVector(0,0,0);
Mesh.Verts[v].Colour := TColorToTVector3f(RGB(Random(256),Random(256),Random(256)));

For W := 0 to Mesh.Verts[v].WeightCount -1 do
begin
Joint := MD5.Joints[Mesh.Weights[Mesh.Verts[v].WeightId + W].BoneId].Position;
TM := MD5.Joints[Mesh.Weights[Mesh.Verts[v].WeightId + W].BoneId].Matrix;
Weighting := Mesh.Weights[Mesh.Verts[v].WeightId + W].Factor;

Pos := Mesh.Weights[Mesh.Verts[v].WeightId + W].Position;
Pos := GetPosWithMatrix(Pos,TM);
Pos := AddVector(Pos,Joint);
Pos := ScaleVector(Pos,Weighting);

Mesh.Verts[v].Position := AddVector(Mesh.Verts[v].Position,Pos);

end;
end;


end;

Procedure BuildMeshs(Var MD5 : TMD5File);
var
X : Integer;
begin

for x := 0 to MD5.Header.NumMeshes-1 do
BuildMD5Mesh(MD5,MD5.Meshs[x]);

end;

Procedure RenderMD5Mesh(Var Mesh : TMD5Mesh);
var
X : integer;
Vertex : Array [0..2] of TBSPVertex;
begin

If Mesh.ShaderID = -2 then
begin
Mesh.ShaderID := ShaderManager.GetShaderId(Mesh.Shader);
MessageBox(0,pchar(Mesh.Shader),'hmm',0);
end;

If Mesh.ShaderID < 0  then
glBindTexture(GL_TEXTURE_2D,0);

for x := 0 to Mesh.Header.NumTris-1 do
begin

If Mesh.ShaderID < 0 then
begin

glbegin(GL_TRIANGLES);
With Mesh.Verts[Mesh.Tris[x].V1] do
glColor3f(Colour.X,Colour.Y,Colour.Z);

With Mesh.Verts[Mesh.Tris[x].V1] do
glVertex3f(Position.X,Position.Z,-Position.Y);

With Mesh.Verts[Mesh.Tris[x].V2] do
glColor3f(Colour.X,Colour.Y,Colour.Z);

With Mesh.Verts[Mesh.Tris[x].V2] do
glVertex3f(Position.X,Position.Z,-Position.Y);

With Mesh.Verts[Mesh.Tris[x].V3] do
glColor3f(Colour.X,Colour.Y,Colour.Z);

With Mesh.Verts[Mesh.Tris[x].V3] do
glVertex3f(Position.X,Position.Z,-Position.Y);
glend;

end
else
begin

With Mesh.Verts[Mesh.Tris[x].V1] do
begin
Vertex[0].Position.X := Position.X;
Vertex[0].Position.Z := -Position.Y;
Vertex[0].Position.Y := Position.Z;
Vertex[0].TextureCoord.X := s;
Vertex[0].TextureCoord.Y := t;
Vertex[0].Color[0] := 1;
Vertex[0].Color[1] := 1;
Vertex[0].Color[2] := 1;
Vertex[0].Color[3] := 1;
end;

With Mesh.Verts[Mesh.Tris[x].V2] do
begin
Vertex[1].Position.X := Position.X;
Vertex[1].Position.Z := -Position.Y;
Vertex[1].Position.Y := Position.Z;
Vertex[1].TextureCoord.X := s;
Vertex[1].TextureCoord.Y := t;
Vertex[1].Color[0] := 1;
Vertex[1].Color[1] := 1;
Vertex[1].Color[2] := 1;
Vertex[1].Color[3] := 1;
end;

With Mesh.Verts[Mesh.Tris[x].V3] do
begin
Vertex[2].Position.X := Position.X;
Vertex[2].Position.Z := -Position.Y;
Vertex[2].Position.Y := Position.Z;
Vertex[2].TextureCoord.X := s;
Vertex[2].TextureCoord.Y := t;
Vertex[2].Color[0] := 1;
Vertex[2].Color[1] := 1;
Vertex[2].Color[2] := 1;
Vertex[2].Color[3] := 1;
end;

ShaderManager.DrawFaces(Vertex,3,Mesh.ShaderID);
end;

end;

end;

Procedure RenderMD5(Var MD5 : TMD5File);
var
x : integer;
begin

For x := 0 to MD5.Header.NumMeshes-1 do
RenderMD5Mesh(MD5.Meshs[x])

end;

end.
